/*
 * Device-independent low level convblit routines - framebuffer blits
 *
 * Copyright (c) 2010 Greg Haerr <greg@censoft.com>
 *
 * This file will need to be modified when adding a new hardware framebuffer
 * image format.
 *
 * Currently, 32bpp BGRA, 32bpp RGBA, 24bpp BGR, and 16bpp RGB565/555 are defined.
 *
 * These routines do no range checking, clipping, or cursor
 * overwriting checks, but instead draw directly to the
 * data_out memory buffer specified in the passed BLITPARMS struct.
 */
#include "device.h"
#include "convblit.h"
#include "../drivers/fb.h"    // DRAWON macro

/* for convenience in specifying inline parms*/
#define R   0   /* RGBA parms*/
#define G   1
#define B   2
#define A   3

#define NONE  MWPORTRAIT_NONE
#define LEFT  MWPORTRAIT_LEFT
#define RIGHT MWPORTRAIT_RIGHT
#define DOWN  MWPORTRAIT_DOWN

#define COPY  0   /* mode parm*/
#define SRCOVER 1

/* framebuffer pixel format blit - must handle backwards copy, nonstd rotation code*/
static inline void ALWAYS_INLINE frameblit_blit(PSD psd, PMWBLITPARMS gc)
{
  int op = gc->op;
  int src_pitch, dst_pitch;
  int width, height;
  unsigned char *src, *dst;

  src_pitch = gc->src_pitch;
  dst_pitch = gc->dst_pitch;  /* dst: next line down*/

  src = ((unsigned char *)gc->data)     + gc->srcy * gc->src_pitch + gc->srcx * 4;
  dst = ((unsigned char *)gc->data_out) + gc->dsty * gc->dst_pitch + gc->dstx * 4;

  width = gc->width;
  height = gc->height;

  /* check for backwards copy if dst in src rect, in same psd*/
  //
  // BUGBUG: For performance, we don't consider same psd
  //
  ASSERT (gc->data != gc->data_out);
  if (gc->data == gc->data_out) {
    return ;
  }

  /*
   * NOTE: The default implementation uses APPLYOP() which forces a
   * switch() within the inner loop to select the rop code.
   * A fast implementation of MWROP_COPY is provided for speed.
   * Any other rop can be sped up by including a case for it, and
   * calling APPLYOP with a constant op parameter.
   *
   * The SRC_OVER case must be handled seperately, as APPLYOP doesn't
   * handle it, along with the FIXME other compositing Porter-Duff ops.
   */
  DRAWON;
  switch (op) {
  case MWROP_COPY:
    ASSERT (gc->srcpsd->data_format == MWIF_BGRX8888 || psd == &scrdev);

    /* fast copy implementation, almost identical to default case below*/
    while (--height >= 0)
    {
      register unsigned char *d = dst;
      register unsigned char *s = src;
      register int w = width;

      while (--w >= 0)
      {
        *(ADDR32)d = *(ADDR32)s;
        d += 4;
        s += 4;
      }
      src += src_pitch;
      dst += dst_pitch;
    }
    break;

  case MWROP_SRC_OVER:
    /* src_over supported for 32bpp framebuffer only*/
    ASSERT (gc->srcpsd->data_format == MWIF_BGRA8888);

    /* src_over only supported on 32bpp framebuffer*/
    while (--height >= 0)
    {
      register unsigned char *s = src;
      register unsigned char *d = dst;
      register int w = width;

      while (--w >= 0)
      {
        register unsigned int alpha;

        alpha = s[3];
        if (alpha == 255)       /* copy source*/
        {
          *(ADDR32)d = *(ADDR32)s;
        }
        else if (alpha != 0)          /* blend source w/dest*/
        {
          d[0] += muldiv255(alpha, s[0] - d[0]);
          d[1] += muldiv255(alpha, s[1] - d[1]);
          d[2] += muldiv255(alpha, s[2] - d[2]);
          d[3] += muldiv255(alpha, 255 - d[3]);
        }
        d += 4;
        s += 4;
      }
      src += src_pitch;
      dst += dst_pitch;
    }
    break;

  default:
    ASSERT (FALSE);
  }
  DRAWOFF;

  /* update screen bits if driver requires it*/
  if (!psd->Update)
    return;

  psd->Update(psd, gc->dstx, gc->dsty, gc->width, gc->height);
}

/* framebuffer pixel format copy blit - 32bpp*/
void frameblit_xxxa8888(PSD psd, PMWBLITPARMS gc)
{
  /* NOTE: src_over works for alpha in fourth byte only (RGBA and BGRA)*/
  frameblit_blit(psd, gc);
}

/* framebuffer pixel format stretch blit - src/dst rotation code, no backwards copy*/
static inline void ALWAYS_INLINE frameblit_stretchblit(PSD psd, PMWBLITPARMS gc)
{
  int src_x_step = gc->src_x_step;    /* normal steps in source image*/
  int src_y_step = gc->src_y_step;
  int src_x_step_one = gc->src_x_step_one; /* 1-unit steps in source image*/
  int src_y_step_one = gc->src_y_step_one;
  int err_x_step = gc->err_x_step;    /* 1-unit error steps in source image*/
  int err_y_step = gc->err_y_step;
  int err_y = gc->err_y;          /* source coordinate error tracking*/
  int err_x_start = gc->err_x;
  int dst_y_step;             /* normal steps in dest image*/
  int x_denominator = gc->x_denominator;
  int y_denominator = gc->y_denominator;
  int op = gc->op;
  int width = gc->width;
  int height = gc->height;
  int src_pitch, dst_pitch;       /* outer loop step*/
  unsigned char * src;      /* source image ptr*/
  unsigned char * dst;      /* dest image ptr*/


  src_pitch = gc->src_pitch;
  dst_pitch = gc->dst_pitch;  /* dst: next line down*/


  /* adjust step values based on bytes per pixel and pitch*/
  src_x_step    *= 4;
  src_x_step_one  *= 4;
  src_y_step    *= src_pitch;
  src_y_step_one  *= src_pitch;
  dst_y_step     = dst_pitch;

  src = ((unsigned char *)gc->data)     + gc->srcy * gc->src_pitch + gc->srcx * 4;
  dst = ((unsigned char *)gc->data_out) + gc->dsty * gc->dst_pitch + gc->dstx * 4;

  /*
   * NOTE: The default implementation uses APPLYOP() which forces a
   * switch() within the inner loop to select the rop code.
   * A fast implementation of MWROP_COPY is provided for speed.
   * Both MWROP_COPY and MWROP_SRCOVER are enhanced to convert
   * from 32/24/16bpp src to downsized destinations.
   *
   * Any other rop can be sped up by including a case for it, and
   * calling APPLYOP with a constant op parameter.  The MWROP_XOR_FGBG
   * example shows this.  A specialized very fast example of
   * MWROP_CLEAR is also included.
   *
   * The SRC_OVER case must be handled seperately, as APPLYOP doesn't
   * handle it, along with the other compositing Porter-Duff ops. FIXME
   */
  DRAWON;
  switch (op) {
  case MWROP_COPY:
//printf("sblit copy\n");
    /* fast copy implementation, converts from different DSZ/SSZ sizes*/
    while (--height >= 0)
    {
      register unsigned char *s = src;
      register unsigned char *d = dst;
      int err_x = err_x_start;
      register int w = width;

      while (--w >= 0)
      {
        *(ADDR32)d = *(ADDR32)s;
        d += 4;
        s += src_x_step;

        err_x += err_x_step;
        if (err_x >= 0) {
          s += src_x_step_one;
          err_x -= x_denominator;
        }
      }
      dst += dst_y_step;
      src += src_y_step;

      err_y += err_y_step;
      if (err_y >= 0) {
        src += src_y_step_one;
        err_y -= y_denominator;
      }
    }
    break;

  case MWROP_SRC_OVER:
//printf("sblit src_over\n");
    while (--height >= 0)
    {
      register unsigned char *s = src;
      register unsigned char *d = dst;
      int err_x = err_x_start;
      register int w = width;

      while (--w >= 0)
      {
        register unsigned int alpha;

        alpha = s[3];
        if (alpha == 255)       /* copy source*/
        {
          *(ADDR32)d = *(ADDR32)s;
        }
        else if (alpha != 0)          /* blend source w/dest*/
        {
          /* d += muldiv255(a, s - d)*/
          d[0] += muldiv255(alpha, s[0] - d[0]);
          d[1] += muldiv255(alpha, s[1] - d[1]);
          d[2] += muldiv255(alpha, s[2] - d[2]);

          d[3] += muldiv255(alpha, 255 - d[3]);
        }
        d += 4;
        s += src_x_step;

        err_x += err_x_step;
        if (err_x >= 0) {
          s += src_x_step_one;
          err_x -= x_denominator;
        }
      }
      dst += dst_y_step;
      src += src_y_step;

      err_y += err_y_step;
      if (err_y >= 0) {
        src += src_y_step_one;
        err_y -= y_denominator;
      }
    }
    break;

  default:
    ASSERT (FALSE);
  }
  DRAWOFF;

  /* update screen bits if driver requires it*/
  if (!psd->Update)
    return;

  psd->Update(psd, gc->dstx, gc->dsty, gc->width, gc->height);
}

/* framebuffer pixel format stretch blit - 32bpp with alpha in 4th byte*/
void
frameblit_stretch_xxxa8888(PSD psd, PMWBLITPARMS gc)
{
  /* NOTE: src_copy works for alpha in fourth byte only (RGBA and BGRA)*/
  frameblit_stretchblit(psd, gc);
}

